from functools import partial
from typing import Sequence, Union

import drawSvg
import numpy as np
from drawSvg import DrawingBasicElement as Drawable
from multimethod import multimethod

from .utils import unwrap_primitives
from .primitives import Line, Polygon, Rectangle, Paste, Circle
from .widget import Widget

try:
    from pyvips import Image

    error_message = None
except OSError as e:
    if 'libvips.so' in str(e):
        error_message = 'The package `libvips` is not installed in your OS'
    elif 'libgobject-2.0.so.0' in str(e):
        error_message = (
            'Your pyvips installation is broken. To fix it type: '
            'pip uninstall -y pyvips && conda install --channel conda-forge pyvips'
        )
    else:
        raise


    class NoStatic(type):
        def __getattr__(self, item):
            if item in ['__origin__', '__args__']:
                raise AttributeError(item)
            raise OSError(error_message)


    class Image(metaclass=NoStatic):
        def __init__(self):
            raise OSError(error_message)


class Vips:
    def __init__(self):
        if error_message:
            raise OSError(error_message)

    @multimethod
    def to_svg(self, image, primitive) -> Drawable:
        raise TypeError(f"Cannot to_svg argument of type {type(primitive)}")

    @to_svg.register
    def to_svg(self, image, primitive: Line) -> Drawable:
        return drawSvg.Line(
            # TODO: add shifts?
            primitive.start[0], image.height - primitive.start[1],
            primitive.stop[0], image.height - primitive.stop[1],
            stroke=primitive.stroke_color.hex, stroke_width=primitive.stroke_width,
            stroke_dasharray=' '.join(map(str, primitive.dash)),
            fill='none', fill_opacity=0.0,
        )

    @to_svg.register
    def to_svg(self, image, primitive: Circle) -> Drawable:
        return drawSvg.Circle(
            cx=primitive.center[0], cy=image.height - primitive.center[1], r=primitive.radius,
            stroke=primitive.stroke_color.hex, stroke_width=primitive.stroke_width,
            stroke_dasharray=' '.join(map(str, primitive.dash)),
            fill='none', fill_opacity=0.0,
        )

    @to_svg.register
    def to_svg(self, image, primitive: Rectangle) -> Drawable:
        delta = primitive.stroke_width * primitive.stroke_location.value
        half_delta = delta * 0.5
        draw_rectangle = partial(
            drawSvg.Rectangle,
            x=primitive.left - half_delta, y=image.height - primitive.height - primitive.top - half_delta,
            width=primitive.width + delta, height=primitive.height + delta,
            stroke=primitive.stroke_color.hex, stroke_width=primitive.stroke_width,
            stroke_dasharray=' '.join(map(str, primitive.dash)),
            fill=primitive.background_color.hex, fill_opacity=primitive.background_color.a / 255,
        )
        if primitive.radius:
            return draw_rectangle(rx=primitive.radius, ry=primitive.radius)
        return draw_rectangle()

    @to_svg.register
    def to_svg(self, image, primitive: Polygon) -> Drawable:
        points_flat = []
        for point in primitive.points:
            points_flat.extend([point[0] + 0.5, image.height - point[1] + 0.5])

        return drawSvg.Lines(
            *points_flat, close=True,
            stroke=primitive.stroke_color.hex, stroke_width=primitive.stroke_width,
            stroke_dasharray=' '.join(map(str, primitive.dash)),
            fill=primitive.background_color.hex, fill_opacity=primitive.background_color.a / 255,
        )

    @multimethod
    def render(self, image, primitive):
        raise TypeError(f"Cannot render argument of type {type(primitive)}")

    @render.register
    def render(self, image, primitive: Union[Line, Rectangle, Polygon, Circle]) -> Image:
        drawing = drawSvg.Drawing(image.width, image.height)
        drawing.append(self.to_svg(image, primitive))
        return paste(image, Image.svgload_buffer(drawing.asSvg().encode()), 0, 0)

    @render.register
    def render(self, image, primitive: Paste) -> Image:
        return paste(image, primitive.patch, primitive.top, primitive.left)

    def draw(self, image, widgets: Sequence[Widget]) -> np.ndarray:
        shape = image.shape[1], image.shape[0]
        image = self.wrap(image)
        layouts = [w.build().align(*shape) for w in widgets]
        for widget, layout in zip(widgets, layouts, strict=True):
            for primitive in unwrap_primitives(widget, layout, int(layout.x), int(layout.y)):
                image = self.render(image, primitive)

        return self.encode(image)

    @staticmethod
    def wrap(image: np.ndarray) -> Image:
        assert image.dtype == np.uint8
        assert image.ndim in [2, 3]
        if image.ndim == 2:
            image = np.repeat(image[..., None], 3, axis=-1)
        else:
            assert image.shape[-1] in [3, 4], image.shape

        height, width, bands = image.shape
        linear = image.reshape(width * height * bands)
        return Image.new_from_memory(linear.data, width, height, bands, 'uchar').copy(interpretation='srgb')

    @staticmethod
    def encode(image):
        result = np.ndarray(
            buffer=image.write_to_memory(), dtype=np.uint8, shape=[image.height, image.width, image.bands]
        )[..., :3]
        # assert result.shape == (image.height, image.width, 3)
        assert result.dtype == np.uint8
        return result


def paste(image: Image, patch: np.ndarray | Image, top, left) -> Image:
    with patch if isinstance(patch, Image) else Vips.wrap(patch) as patch:
        return image.composite(patch, 'over', x=left, y=top)



from enum import Enum
from typing import NamedTuple, Union, Sequence, Type, Tuple, Any


class Auto:
    def __init__(self):
        raise RuntimeError


number = Union[int, float]


class Percent(NamedTuple):
    value: number


Amount = Union[Type[Auto], Percent, number]
MaybePercent = Union[number, Percent]

MarginLike = Union[Amount, Sequence[Amount]]
BorderLike = Union[number, Sequence[number]]
PaddingLike = Union[MaybePercent, Sequence[MaybePercent]]


class Direction(Enum):
    Row, Column, RowReverse, ColumnReverse = range(4)


class PositionType(Enum):
    Relative, Absolute = range(2)


class Wrap(Enum):
    NoWrap, Wrap, Reverse = range(3)

    @classmethod
    def from_bool(cls, x):
        if isinstance(x, bool):
            return cls.Wrap if x else cls.NoWrap
        return x


class Align(Enum):
    Auto, Start, Center, End, Stretch, Baseline, SpaceBetween, SpaceAround = range(8)

    @classmethod
    def from_auto(cls, x):
        if x is Auto:
            return cls.Auto
        return x


class Justify(Enum):
    Start, Center, End, SpaceBetween, SpaceAround, SpaceEvenly = range(6)


class Flex:
    def __init__(self, min_width: MaybePercent = None, width: Amount = Auto, max_width: MaybePercent = None,
                 min_height: MaybePercent = None, height: Amount = Auto, max_height: MaybePercent = None,
                 margin: MarginLike = 0, border: BorderLike = 0, padding: PaddingLike = 0,
                 grow: number = 0, shrink: number = 0, basis: Amount = Auto,
                 direction: Direction = Direction.Row, wrap: Union[Wrap, bool] = False,
                 align_content: Align = Align.Start, align_items: Align = Align.Stretch, align_self: Align = Align.Auto,
                 justify_content: Justify = Justify.Start, position_type: PositionType = PositionType.Relative,
                 aspect_ratio: float = None, children: Sequence['Flex'] = (), context: Any = None):
        # TODO: better values validation
        self.position_type = position_type
        self.justify_content = justify_content
        self.aspect_ratio = aspect_ratio
        self.align_self = Align.from_auto(align_self)
        self.align_items = Align.from_auto(align_items)
        self.align_content = Align.from_auto(align_content)
        self.wrap = Wrap.from_bool(wrap)
        self.padding = _broadcast(padding)
        self.border = _broadcast(border)
        self.margin = _broadcast(margin)
        self.max_height = max_height
        self.max_width = max_width
        self.min_height = min_height
        self.min_width = min_width
        self.direction = direction
        self.basis = basis
        self.width, self.height = width, height
        self.grow, self.shrink = grow, shrink
        self.children = tuple(children)
        self.context = context

    def align(self, width: number, height: number) -> 'Layout':
        from .bridge import align
        return align(self, width, height)


class BoxSides(NamedTuple):
    begin: float
    end: float
    top: float
    bottom: float

    @property
    def horizontal(self):
        return self.begin + self.end

    @property
    def vertical(self):
        return self.top + self.bottom


class Layout(NamedTuple):
    x: float
    y: float
    rx: float
    ry: float
    width: float
    height: float
    margin: BoxSides
    border: BoxSides
    padding: BoxSides
    flex: Flex
    children: Tuple['Layout', ...]

    @property
    def position(self):
        return self.x, self.y

    @property
    def size(self):
        return self.width, self.height

    @property
    def context(self):
        return self.flex.context


def _broadcast(x):
    if isinstance(x, (int, float, Percent)) or x is Auto:
        return (x,) * 4
    return x



from static_yoga import Flex, Justify


def test_align():
    root = Flex(
        justify_content=Justify.SpaceBetween,
        children=[
            Flex(100, 100),
            Flex(100, 100),
        ]
    ).align(1000, 1000)
    a, b = root.children
    assert a.position == (0, 0)
    assert a.size == (100, 1000)
    assert b.position == (900, 0)
    assert b.size == (100, 1000)



from .primitives import Primitive


def unwrap_primitives(widget, layout, shift_x, shift_y):
    for item in widget.draw(layout):
        if isinstance(item, Primitive):
            yield item.shift(int(shift_x), int(shift_y))
        else:
            child_layout, child_widget = item
            yield from unwrap_primitives(
                child_widget, child_layout,
                shift_x + child_layout.rx, shift_y + child_layout.ry,
            )



# cython: language_level=3

from .bridge cimport *
from .interfaces import Flex, Layout, Auto, Percent, Direction, Wrap, Align, Justify, BoxSides, PositionType

ctypedef void (*Amount1)(YGNode *)
ctypedef void (*Amount2)(YGNode *, float)
ctypedef void (*Edge1)(YGNode *, YGEdge)
ctypedef void (*Edge2)(YGNode *, YGEdge, float)
ctypedef void (*AlignFunc)(YGNode *, YGAlign)

cdef inline void _amount(YGNode * node, value, Amount1 on_auto, Amount2 on_percent, Amount2 on_absolute):
    if value is Auto:
        on_auto(node)
    elif isinstance(value, Percent):
        on_percent(node, value.value)
    elif isinstance(value, (int, float)):
        on_absolute(node, value)
    else:
        raise TypeError(value)

cdef inline void _amount_edge(YGNode * node, values, Edge1 on_auto, Edge2 on_percent, Edge2 on_absolute):
    for idx, value in enumerate(values):
        if value is Auto:
            on_auto(node, _edge(idx))
        elif isinstance(value, Percent):
            on_percent(node, _edge(idx), value.value)
        elif isinstance(value, (int, float)):
            on_absolute(node, _edge(idx), value)
        else:
            raise TypeError(value)

cdef inline void _maybe_percent(YGNode * node, value, Amount2 on_percent, Amount2 on_absolute):
    if value is None:
        return
    _amount(node, value, NULL, on_percent, on_absolute)

cdef inline _edge(int idx):
    if idx == 0:
        return YGEdgeStart
    if idx == 1:
        return YGEdgeEnd
    if idx == 2:
        return YGEdgeTop
    if idx == 3:
        return YGEdgeBottom
    raise ValueError(idx)

cdef _align(AlignFunc func, YGNode * node, value):
    if value is Align.Auto:
        func(node, YGAlignAuto)
    elif value is Align.Start:
        func(node, YGAlignFlexStart)
    elif value is Align.Center:
        func(node, YGAlignCenter)
    elif value is Align.End:
        func(node, YGAlignFlexEnd)
    elif value is Align.Stretch:
        func(node, YGAlignStretch)
    elif value is Align.Baseline:
        func(node, YGAlignBaseline)
    elif value is Align.SpaceBetween:
        func(node, YGAlignSpaceBetween)
    elif value is Align.SpaceAround:
        func(node, YGAlignSpaceAround)
    else:
        raise TypeError(value)

cdef YGNode * forward(node: Flex):
    yg_node = new YGNode()

    try:
        # width
        _amount(yg_node, node.width, YGNodeStyleSetWidthAuto, YGNodeStyleSetWidthPercent, YGNodeStyleSetWidth)
        _maybe_percent(yg_node, node.min_width, YGNodeStyleSetMinWidthPercent, YGNodeStyleSetMinWidth)
        _maybe_percent(yg_node, node.max_width, YGNodeStyleSetMaxWidthPercent, YGNodeStyleSetMaxWidth)
        # height
        _amount(yg_node, node.height, YGNodeStyleSetHeightAuto, YGNodeStyleSetHeightPercent, YGNodeStyleSetHeight)
        _maybe_percent(yg_node, node.min_height, YGNodeStyleSetMinHeightPercent, YGNodeStyleSetMinHeight)
        _maybe_percent(yg_node, node.max_height, YGNodeStyleSetMaxHeightPercent, YGNodeStyleSetMaxHeight)
        # margin stuff
        _amount_edge(yg_node, node.margin, YGNodeStyleSetMarginAuto, YGNodeStyleSetMarginPercent, YGNodeStyleSetMargin)
        _amount_edge(yg_node, node.border, NULL, NULL, YGNodeStyleSetBorder)
        _amount_edge(yg_node, node.padding, NULL, YGNodeStyleSetPaddingPercent, YGNodeStyleSetPadding)

        # flex
        if node.direction is Direction.Row:
            YGNodeStyleSetFlexDirection(yg_node, YGFlexDirectionRow)
        elif node.direction is Direction.RowReverse:
            YGNodeStyleSetFlexDirection(yg_node, YGFlexDirectionRowReverse)
        elif node.direction is Direction.Column:
            YGNodeStyleSetFlexDirection(yg_node, YGFlexDirectionColumn)
        elif node.direction is Direction.ColumnReverse:
            YGNodeStyleSetFlexDirection(yg_node, YGFlexDirectionColumnReverse)
        else:
            raise TypeError(node.direction)

        if node.justify_content is Justify.Start:
            YGNodeStyleSetJustifyContent(yg_node, YGJustifyFlexStart)
        elif node.justify_content is Justify.Center:
            YGNodeStyleSetJustifyContent(yg_node, YGJustifyCenter)
        elif node.justify_content is Justify.End:
            YGNodeStyleSetJustifyContent(yg_node, YGJustifyFlexEnd)
        elif node.justify_content is Justify.SpaceBetween:
            YGNodeStyleSetJustifyContent(yg_node, YGJustifySpaceBetween)
        elif node.justify_content is Justify.SpaceAround:
            YGNodeStyleSetJustifyContent(yg_node, YGJustifySpaceAround)
        elif node.justify_content is Justify.SpaceEvenly:
            YGNodeStyleSetJustifyContent(yg_node, YGJustifySpaceEvenly)
        else:
            raise TypeError(node.justify_content)

        # align
        _align(YGNodeStyleSetAlignContent, yg_node, node.align_content)
        _align(YGNodeStyleSetAlignItems, yg_node, node.align_items)
        _align(YGNodeStyleSetAlignSelf, yg_node, node.align_self)

        # position
        if node.position_type is PositionType.Absolute:
            YGNodeStyleSetPositionType(yg_node, YGPositionTypeAbsolute)

        # stretching
        YGNodeStyleSetFlexGrow(yg_node, node.grow)
        YGNodeStyleSetFlexShrink(yg_node, node.shrink)
        _amount(
            yg_node, node.basis, YGNodeStyleSetFlexBasisAuto, YGNodeStyleSetFlexBasisPercent, YGNodeStyleSetFlexBasis
        )

        if node.wrap is Wrap.NoWrap:
            YGNodeStyleSetFlexWrap(yg_node, YGWrapNoWrap)
        elif node.wrap is Wrap.Wrap:
            YGNodeStyleSetFlexWrap(yg_node, YGWrapWrap)
        elif node.wrap is Wrap.Reverse:
            YGNodeStyleSetFlexWrap(yg_node, YGWrapWrapReverse)
        else:
            raise TypeError(node.wrap)

        if node.aspect_ratio is not None:
            YGNodeStyleSetAspectRatio(yg_node, node.aspect_ratio)

        for idx, child in enumerate(node.children):
            YGNodeInsertChild(yg_node, forward(child), idx)

    except:
        YGNodeFree(yg_node)
        raise

    return yg_node

cdef backward(YGNode * internal, external, x, y):
    layout = internal.getLayout()
    local_x, local_y = layout.position[0], layout.position[1]
    global_x, global_y = local_x + x, local_y + y
    return Layout(
        global_x, global_y, local_x, local_y,
        layout.dimensions[0], layout.dimensions[1],
        BoxSides(*(layout.margin[i] for i in range(4))),
        BoxSides(*(layout.border[i] for i in range(4))),
        BoxSides(*(layout.padding[i] for i in range(4))),
        flex=external, children=tuple(
            backward(YGNodeGetChild(internal, idx), child, global_x, global_y)
            for idx, child in enumerate(external.children)
        ),
    )

cpdef align(root: Flex, width: float, height: float):
    try:
        yg_root = forward(root)
        YGNodeCalculateLayout(yg_root, width, height, YGDirectionLTR)
        return backward(yg_root, root, 0, 0)

    finally:
        # gc can't automatically clean these objects
        YGNodeFree(yg_root)



from libc.stdint cimport uint32_t

cdef extern from "<array>" namespace "std" nogil:
    cdef cppclass array2 "std::array<float, 2>":
        array2() except+
        float& operator[](size_t)

    cdef cppclass array4 "std::array<float, 4>":
        array4() except+
        float& operator[](size_t)

cdef extern from "yoga/YGNode.cpp":
    pass

cdef extern from "yoga/YGLayout.cpp":
    pass

cdef extern from "yoga/YGEnums.cpp":
    pass

cdef extern from "yoga/YGLayout.h":
    cdef cppclass YGLayout:
        YGLayout() except +
        array2 dimensions
        array4 position, margin, border, padding

cdef extern from "yoga/YGNode.h":
    cdef cppclass YGNode:
        YGNode() except +
        YGLayout& getLayout()


cdef extern from "yoga/YGEnums.h":
    cdef enum YGDirection:
        YGDirectionInherit,
        YGDirectionLTR,
        YGDirectionRTL

    cdef enum YGEdge:
        YGEdgeLeft,
        YGEdgeTop,
        YGEdgeRight,
        YGEdgeBottom,
        YGEdgeStart,
        YGEdgeEnd,
        YGEdgeHorizontal,
        YGEdgeVertical,
        YGEdgeAll

    cdef enum YGFlexDirection:
        YGFlexDirectionColumn,
        YGFlexDirectionColumnReverse,
        YGFlexDirectionRow,
        YGFlexDirectionRowReverse

    cdef enum YGAlign:
        YGAlignAuto,
        YGAlignFlexStart,
        YGAlignCenter,
        YGAlignFlexEnd,
        YGAlignStretch,
        YGAlignBaseline,
        YGAlignSpaceBetween,
        YGAlignSpaceAround

    cdef enum YGJustify:
        YGJustifyFlexStart,
        YGJustifyCenter,
        YGJustifyFlexEnd,
        YGJustifySpaceBetween,
        YGJustifySpaceAround,
        YGJustifySpaceEvenly

    cdef enum YGOverflow:
        YGOverflowVisible,
        YGOverflowHidden,
        YGOverflowScroll

    cdef enum YGWrap:
        YGWrapNoWrap,
        YGWrapWrap,
        YGWrapWrapReverse

    cdef enum YGPositionType:
        YGPositionTypeStatic,
        YGPositionTypeRelative,
        YGPositionTypeAbsolute

cdef extern from "yoga/Yoga.h":
    # dims
    void YGNodeStyleSetWidth(YGNode * node, float & points)
    void YGNodeStyleSetWidthPercent(YGNode * node, float & percent)
    void YGNodeStyleSetWidthAuto(YGNode * node)
    void YGNodeStyleSetHeight(YGNode * node, float & points)
    void YGNodeStyleSetHeightPercent(YGNode * node, float & percent)
    void YGNodeStyleSetHeightAuto(YGNode * node)

    void YGNodeStyleSetMinWidth(YGNode * node, float minWidth)
    void YGNodeStyleSetMinWidthPercent(YGNode * node, float minWidth)
    void YGNodeStyleSetMinHeight(YGNode * node, float minHeight)
    void YGNodeStyleSetMinHeightPercent(YGNode * node, float minHeight)
    void YGNodeStyleSetMaxWidth(YGNode * node, float maxWidth)
    void YGNodeStyleSetMaxWidthPercent(YGNode * node, float maxWidth)
    void YGNodeStyleSetMaxHeight(YGNode * node, float maxHeight)
    void YGNodeStyleSetMaxHeightPercent(YGNode * node, float maxHeight)

    void YGNodeStyleSetMargin(YGNode * node, YGEdge edge, float margin)
    void YGNodeStyleSetMarginPercent(YGNode * node, YGEdge edge, float margin)
    void YGNodeStyleSetMarginAuto(YGNode * node, YGEdge edge)
    void YGNodeStyleSetPadding(YGNode * node, YGEdge edge, float padding)
    void YGNodeStyleSetPaddingPercent(YGNode * node, YGEdge edge, float padding)
    void YGNodeStyleSetBorder(YGNode * node, YGEdge edge, float border)

    # flex
    void YGNodeStyleSetFlexDirection(YGNode * node, YGFlexDirection flexDirection)
    void YGNodeStyleSetJustifyContent(YGNode * node, YGJustify justifyContent)
    void YGNodeStyleSetAlignContent(YGNode * node, YGAlign alignContent)
    void YGNodeStyleSetAlignItems(YGNode * node, YGAlign alignItems)
    void YGNodeStyleSetAlignSelf(YGNode * node, YGAlign alignSelf)
    void YGNodeStyleSetPositionType(YGNode * node, YGPositionType positionType)
    void YGNodeStyleSetFlexWrap(YGNode * node, YGWrap flexWrap)
    # void YGNodeStyleSetOverflow(YGNode * node, YGOverflow overflow)
    # void YGNodeStyleSetDisplay(YGNode * node, YGDisplay display)
    # void YGNodeStyleSetFlex(YGNode * node, float flex)
    void YGNodeStyleSetFlexGrow(YGNode * node, float flexGrow)
    void YGNodeStyleSetFlexShrink(YGNode * node, float flexShrink)
    void YGNodeStyleSetFlexBasis(YGNode * node, float flexBasis)
    void YGNodeStyleSetFlexBasisPercent(YGNode * node, float flexBasis)
    void YGNodeStyleSetFlexBasisAuto(YGNode * node)
    void YGNodeStyleSetAspectRatio(YGNode * node, float aspectRatio)

    # structure
    void YGNodeInsertChild(YGNode * owner, YGNode * child, uint32_t index)
    void YGNodeFree(YGNode * node)
    YGNode * YGNodeGetChild(YGNode * node, uint32_t index)
    void YGNodeCalculateLayout(YGNode * node, float ownerWidth, float ownerHeight, YGDirection ownerDirection)



from pathlib import Path
from subprocess import check_call
from tempfile import TemporaryDirectory

from setuptools import Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py

ROOT = Path(__file__).parent.resolve()
name = 'static_yoga'
YOGA_ROOT = ROOT / name / 'yoga' / 'yoga'
# BUILD_ROOT = None
# useful for caching the compilation
BUILD_ROOT = ROOT / 'build' / 'yoga'


class PyprojectBuild(build_py):
    def initialize_options(self):
        super().initialize_options()

        self.distribution.ext_modules = [Extension(
            f'{name}.yoga.bridge',
            sources=[f'{name}/yoga/bridge.pyx'],
            include_dirs=[str(YOGA_ROOT)],
            language='c++',
            extra_compile_args=['-std=c++11'],
        )]


class YogaBuilder(build_ext):
    def build_extension(self, ext):
        with TemporaryDirectory() as _root:
            if BUILD_ROOT is not None:
                BUILD_ROOT.mkdir(parents=True, exist_ok=True)
                root = BUILD_ROOT
            else:
                root = Path(_root)

            # build
            check_call(['cmake', '-S', str(YOGA_ROOT), f'-B{root}'])
            # compile
            check_call(['make'], cwd=root)

            ext.extra_objects.append(str(root / 'libyogacore.a'))
            self.compiler.add_include_dir(str(YOGA_ROOT / 'yoga'))
            super().build_extension(ext)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable, Sequence

from .yoga import Flex, Layout, Percent
from .primitives import Primitive


class Widget(ABC):
    """
    Basic block for all UI elements. Knows how to position and draw itself given a layout
    """

    @abstractmethod
    def build(self) -> Flex:
        """ Return a layout box """

    @abstractmethod
    def draw(self, layout: Layout) -> Iterable[Drawable]:
        """ Yield primitive draw calls which will be used to render the widget """


Drawable = Primitive | tuple[Layout, Widget]


class PrimitiveWidget(Widget):
    def __init__(self, primitive: Primitive):
        self.primitive = primitive

    def build(self) -> Flex:
        return Flex(width=Percent(100), height=Percent(100))

    def draw(self, layout: Layout) -> Iterable[Drawable]:
        yield self.primitive

    @staticmethod
    def wrap(*instances) -> Sequence[Widget]:
        return tuple(PrimitiveWidget(x) if isinstance(x, Primitive) else x for x in instances)


class WidgetFactory(Widget):
    def __init__(self, widget: Widget):
        super().__init__()
        self._widget = widget

    def build(self) -> Flex:
        return self._widget.build()

    def draw(self, layout: Layout) -> Iterable[Drawable]:
        return self._widget.draw(layout)



from abc import ABC, abstractmethod
from enum import Enum
from typing import Sequence, Tuple

import numpy as np
from pydantic import BaseModel, Extra

from .color import Color


class NoExtra(BaseModel):
    class Config:
        extra = Extra.forbid


class Primitive(ABC):
    """ Base class for all primitive drawing operations """

    @abstractmethod
    def shift(self, x: float, y: float):
        """ Return primitives with shifted coordinates"""


class StrokeLocation(Enum):
    """ Defines how the stroke is drawn """
    Inner, Central, Outer = range(-1, 2)


class Paste(Primitive, NoExtra):
    patch: np.ndarray
    top: int
    left: int

    class Config:
        arbitrary_types_allowed = True

    def shift(self, x, y):
        return self.copy(update={'top': y + self.top, 'left': x + self.left})


class Line(Primitive, NoExtra):
    start: Tuple[float, float]
    stop: Tuple[float, float]
    stroke_color: Color
    stroke_width: float
    dash: tuple = ()

    def shift(self, x, y):
        return self.copy(update={'start': (self.start[0] + x, self.start[1] + y),
                                 'stop': (self.stop[0] + x, self.stop[1] + y)})


class Circle(Primitive, NoExtra):
    center: Tuple[float, float]
    radius: float
    stroke_color: Color
    stroke_width: float
    dash: tuple = ()

    def shift(self, x, y):
        return self.copy(update={'center': (self.center[0] + x, self.center[1] + y)})


class Polygon(Primitive, NoExtra):
    points: Sequence[Tuple[float, float]]
    stroke_color: Color = Color.Transparent
    stroke_width: float = 0
    background_color: Color = Color.Transparent
    dash: tuple = ()

    def shift(self, x, y):
        return self.copy(update={'points': [(p[0] + x, p[1] + y) for p in self.points]})


class Rectangle(Primitive, NoExtra):
    top: float
    left: float
    width: float
    height: float
    stroke_color: Color = Color.Transparent
    stroke_width: float = 0
    stroke_location: StrokeLocation = StrokeLocation.Inner
    background_color: Color = Color.Transparent
    radius: float = 0
    dash: Tuple[float, ...] = ()

    def shift(self, x, y):
        return self.copy(update={'top': self.top + y, 'left': self.left + x})



from typing import Sequence, Union

import numpy as np
from multimethod import multimethod
from PIL.ImageColor import getrgb

__all__ = 'Color', 'ColorLike'


class Color:
    r: int
    g: int
    b: int
    a: int

    @multimethod
    def __init__(self, r: Union[int, np.uint8], g: Union[int, np.uint8], b: Union[int, np.uint8],
                 a: Union[int, np.uint8] = 255):
        super().__init__()
        self.r, self.g, self.b, self.a = int(r), int(g), int(b), int(a)

    @__init__.register
    def __init__(self, rgba: Sequence[Union[int, np.uint8]]):
        self.__init__(*rgba)

    @__init__.register
    def __init__(self, rgb_or_name: str):
        self.__init__(*getrgb(rgb_or_name))

    @__init__.register
    def __init__(self, rgba: int):
        self.__init__(hex(rgba).replace('0x', '#'))

    @__init__.register
    def __init__(self, color: 'Color'):
        self.__init__(color.rgba)

    @property
    def rgb(self):
        return self.r, self.g, self.b

    @property
    def rgb_unit(self):
        return self.r / 255, self.g / 255, self.b / 255

    @property
    def rgba_unit(self):
        return self.r / 255, self.g / 255, self.b / 255, self.a / 255

    @property
    def rgba(self):
        return self.r, self.g, self.b, self.a

    @property
    def hex(self):
        components = self.rgba
        if components[-1] == 255:
            components = components[:-1]
        return '#' + ''.join(f'{x:02x}' for x in components)

    def __eq__(self, other: 'Color'):
        return (
            self.r == other.r and self.g == other.g and self.b == other.b and self.a == other.a
        )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        return cls(v)

    def __repr__(self):
        return f'<Color {self.rgba}>'

    # presets
    Transparent: 'Color' = 0, 0, 0, 0
    Red: 'Color' = 255, 0, 0, 255
    Green: 'Color' = 0, 255, 0, 255
    Gray: 'Color' = 128, 128, 128, 255
    DarkGray: 'Color' = 64, 64, 64, 255
    ForestGreen: 'Color' = 34, 139, 34, 255
    Yellow: 'Color' = 255, 255, 0, 255
    Orange: 'Color' = 255, 165, 0, 255
    DarkOrange: 'Color' = 255, 140, 0, 255
    Blue: 'Color' = 0, 0, 255, 255
    Black: 'Color' = 0, 0, 0, 255
    White: 'Color' = 255, 255, 255, 255


for _attr, _kind in Color.__annotations__.items():
    if _kind == 'Color' and not isinstance(getattr(Color, _attr), Color):
        setattr(Color, _attr, Color(getattr(Color, _attr)))

ColorLike = Union[Color, str, int, Sequence[int]]
